"""
Provides FileStorage implementation for MongoDB GridFS.

This is useful for storing files inside a mongodb database.

"""
from __future__ import absolute_import

from datetime import datetime
import uuid
from boto.s3.connection import S3Connection
from depot._compat import unicode_text

from .interfaces import FileStorage, StoredFile
from . import utils

CANNED_ACL_PUBLIC_READ = 'public-read'
CANNED_ACL_PRIVATE = 'private'


class S3StoredFile(StoredFile):
    def __init__(self, file_id, key):
        _check_file_id(file_id)
        self._key = key

        metadata_info = {'filename': key.get_metadata('x-depot-filename'),
                         'content_type': key.content_type,
                         'content_length': key.size,
                         'last_modified': None}

        try:
            last_modified = key.get_metadata('x-depot-modified')
            if last_modified:
                metadata_info['last_modified'] = datetime.strptime(last_modified,
                                                                   '%Y-%m-%d %H:%M:%S')
        except:
            pass

        super(S3StoredFile, self).__init__(file_id=file_id, **metadata_info)

    def read(self, n=-1):
        if self.closed:
            raise ValueError("cannot read from a closed file")

        return self._key.read(n)

    def close(self):
        self._key.close()

    @property
    def closed(self):
        return self._key.closed

    @property
    def public_url(self):
        return self._key.generate_url(expires_in=0, query_auth=False)


class S3Storage(FileStorage):
    """:class:`depot.io.interfaces.FileStorage` implementation that stores files on S3.

    All the files are stored inside a bucket named ``bucket`` on ``host`` which Depot
    connects to using ``access_key_id`` and ``secret_access_key``. If ``host`` is
    omitted the Amazon AWS S3 storage is used. Additionally, a canned ACL policy of
    either ``private`` or ``public-read`` can be specified with the ``policy`` parameter.
    Finally, the ``encrypt_key`` parameter can be specified to use the server side
    encryption feature.
    """

    def __init__(self, access_key_id, secret_access_key, bucket=None, host=None,
                 policy=None, encrypt_key=False):
        policy = policy or CANNED_ACL_PUBLIC_READ
        assert policy in [CANNED_ACL_PUBLIC_READ, CANNED_ACL_PRIVATE], (
            "Key policy must be %s or %s" % (CANNED_ACL_PUBLIC_READ, CANNED_ACL_PRIVATE))
        self._policy = policy or CANNED_ACL_PUBLIC_READ
        self._encrypt_key = encrypt_key

        if bucket is None:
            bucket = 'filedepot-%s' % (access_key_id.lower(),)

        kw = {}
        if host is not None:
            kw['host'] = host
        self._conn = S3Connection(access_key_id, secret_access_key, **kw)
        self._bucket = self._conn.lookup(bucket)
        if self._bucket is None:
            self._bucket = self._conn.create_bucket(bucket)

    def get(self, file_or_id):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        key = self._bucket.get_key(fileid)
        if key is None:
            raise IOError('File %s not existing' % fileid)

        return S3StoredFile(fileid, key)

    def __save_file(self, key, content, filename, content_type=None):
        key.set_metadata('content-type', content_type)
        key.set_metadata('x-depot-filename', filename)
        key.set_metadata('x-depot-modified', utils.timestamp())
        key.set_metadata('Content-Disposition', 'inline; filename="%s"' % filename)

        if hasattr(content, 'read'):
            can_seek_and_tell = True
            try:
                pos = content.tell()
                content.seek(pos)
            except:
                can_seek_and_tell = False

            if can_seek_and_tell:
                key.set_contents_from_file(content, policy=self._policy,
                                           encrypt_key=self._encrypt_key)
            else:
                key.set_contents_from_string(content.read(), policy=self._policy,
                                             encrypt_key=self._encrypt_key)
        else:
            if isinstance(content, unicode_text):
                raise TypeError('Only bytes can be stored, not unicode')
            key.set_contents_from_string(content, policy=self._policy,
                                         encrypt_key=self._encrypt_key)

    def create(self, content, filename=None, content_type=None):
        content, filename, content_type = self.fileinfo(content, filename, content_type)
        new_file_id = str(uuid.uuid1())
        key = self._bucket.new_key(new_file_id)
        self.__save_file(key, content, filename, content_type)
        return new_file_id

    def replace(self, file_or_id, content, filename=None, content_type=None):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        content, filename, content_type = self.fileinfo(content, filename, content_type)
        if filename is None:
            f = self.get(fileid)
            filename = f.filename
            content_type = f.content_type

        key = self._bucket.get_key(fileid)
        self.__save_file(key, content, filename, content_type)
        return fileid

    def delete(self, file_or_id):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        k = self._bucket.get_key(fileid)
        if k:
            k.delete()

    def exists(self, file_or_id):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        k = self._bucket.get_key(fileid)
        return k is not None


def _check_file_id(file_id):
    # Check that the given file id is valid, this also
    # prevents unsafe paths.
    try:
        uuid.UUID('{%s}' % file_id)
    except:
        raise ValueError('Invalid file id %s' % file_id)



import cgi
from datetime import datetime
from tempfile import SpooledTemporaryFile
from depot._compat import byte_string


INMEMORY_FILESIZE = 1024*1024


def timestamp():
    return datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')


def file_from_content(content):
    """Provides a real file object from file content

    Converts ``FileStorage``, ``FileIntent`` and
    ``bytes`` to an actual file.
    """
    f = content
    if isinstance(content, cgi.FieldStorage):
        f = content.file
    elif isinstance(content, FileIntent):
        f = content._fileobj
    elif isinstance(content, byte_string):
        f = SpooledTemporaryFile(INMEMORY_FILESIZE)
        f.write(content)
    f.seek(0)
    return f


class FileIntent(object):
    """Represents the intention to upload a file

    Whenever a file can be stored by depot, a FileIntent
    can be passed instead of the file itself. This permits
    to easily upload objects that are not files or to add
    missing information to the uploaded files.
    """
    def __init__(self, fileobj, filename, content_type):
        self._fileobj = fileobj
        self._filename = filename
        self._content_type = content_type

    @property
    def fileinfo(self):
        return self._fileobj, self._filename, self._content_type


from depot.io import utils
from depot.manager import DepotManager
from ..interfaces import FileFilter
from PIL import Image
from io import BytesIO


class WithThumbnailFilter(FileFilter):
    def __init__(self, size=(128,128), format='PNG'):
        self.thumbnail_size = size
        self.thumbnail_format = format

    def on_save(self, uploaded_file):
        content = utils.file_from_content(uploaded_file.original_content)

        thumbnail = Image.open(content)
        thumbnail.thumbnail(self.thumbnail_size, Image.BILINEAR)
        thumbnail = thumbnail.convert('RGBA')
        thumbnail.format = self.thumbnail_format

        output = BytesIO()
        thumbnail.save(output, self.thumbnail_format)
        output.seek(0)

        thumb_file_name = 'thumb.%s' % self.thumbnail_format.lower()
        thumb_path, thumb_id = uploaded_file.store_content(output, thumb_file_name)
        uploaded_file['thumb_id'] = thumb_id
        uploaded_file['thumb_path'] = thumb_path
        uploaded_file['thumb_url'] = DepotManager.get_middleware().url_for(thumb_path)





from __future__ import absolute_import

from ming.odm import mapper
from ming.odm.property import FieldProperty
from ming.odm.base import session, state, ObjectState
from ming.odm.odmsession import SessionExtension
from ming.schema import Anything

from depot.manager import DepotManager
from .upload import UploadedFile


class _UploadedFileSchema(Anything):
    pass


class UploadedFileProperty(FieldProperty):
        def __init__(self,  filters=tuple(), upload_type=UploadedFile, upload_storage=None):
            FieldProperty.__init__(self, _UploadedFileSchema())
            self._filters = filters
            self._upload_type = upload_type
            self._upload_storage = upload_storage

        def __set__(self, instance, value):
            if value is not None and not isinstance(value, UploadedFile):
                upload_type = self._upload_type
                value = upload_type(value, self._upload_storage)

            if isinstance(value, UploadedFile):
                value._apply_filters(self._filters)

            old_value = self.__get__(instance, instance.__class__)
            DepotExtension.get_depot_history(instance).swap(old_value, value)
            return FieldProperty.__set__(self, instance, value)

        def __get__(self, instance, owner=None):
            try:
                value = FieldProperty.__get__(self, instance, owner)
            except AttributeError:
                value = None

            if not value:
                return None

            if instance is None:
                return value

            return self._upload_type(value)


        """
        # Looks like this should do nothing on ming.
        def __delete__(self, instance, owner=None):
            old_value = self.__get__(instance, instance.__class__)
            DepotExtension.get_depot_history(instance).delete(old_value)
            return FieldProperty.__delete__(self, instance, owner)
        """


class DepotExtension(SessionExtension):
    @classmethod
    def get_depot_history(cls, instance):
        istate = state(instance)
        if not hasattr(istate, '_depot_history'):
            istate._depot_history = _DepotHistory()
        return istate._depot_history

    def _check_object_deleted(self, obj):
        hist = self.get_depot_history(obj)
        if state(obj).status == ObjectState.deleted:
            for prop in mapper(obj).properties:
                if isinstance(prop, UploadedFileProperty):
                    current_value = prop.__get__(obj, obj.__class__)
                    hist.delete(current_value)
            self._flush_object(obj)

    def _flush_object(self, obj):
        history = self.get_depot_history(obj)
        for entry in history.deleted:
            depot, fileid = entry.split('/', 1)
            depot = DepotManager.get(depot)
            depot.delete(fileid)
        history.clear()

    def before_flush(self, obj=None):
        if obj:
            self._check_object_deleted(obj)
        else:
            for class_, id_, obj in self.session.imap:
                self._check_object_deleted(obj)

    def after_flush(self, obj=None):
        if obj:
            self._flush_object(obj)
        else:
            for class_, id_, obj in self.session.imap:
                self._flush_object(obj)


class _DepotHistory(object):
    def __init__(self):
        self.clear()

    def _extract_files(self, obj):
        return obj['files']

    def add(self, obj):
        if obj is None:
            return

        files = self._extract_files(obj)
        self.deleted.difference_update(obj)
        self.new.update(files)

    def delete(self, obj):
        if obj is None:
            return

        files = self._extract_files(obj)
        self.new.difference_update(obj)
        self.deleted.update(files)

    def swap(self, old, new):
        self.delete(old)
        self.add(new)

    def clear(self):
        self.deleted = set()
        self.new = set()


try:  # pragma: no cover
    from sprox.mg.widgetselector import MingWidgetSelector
    from tw2.forms import FileField as TW2FileField
    MingWidgetSelector.default_widgets.setdefault(_UploadedFileSchema, TW2FileField)

    from sprox.mg.validatorselector import MingValidatorSelector
    from ..validators import TW2FileIntentValidator
    MingValidatorSelector.default_validators.setdefault(_UploadedFileSchema, TW2FileIntentValidator)
except ImportError:  # pragma: no cover
    pass